#property script_show_inputs

input string inFileName = "TP.csv";
input int inMinCount = 1; // .  ,  

#include <fxsaber\TesterCache\String.mqh> // https://www.mql5.com/ru/code/26223

//      .
string TimeToString( const long time, const int FlagTime = TIME_DATE | TIME_SECONDS)
{
  return(TimeToString((datetime)time / 1000, FlagTime) + "." + IntegerToString(time % 1000, 3, '0'));
}

struct ORDER
{
  long Ticket;
  STRING32 Symb;
  ENUM_ORDER_STATE State;
  long TimeSetup;
  long TimeDone;
  ENUM_ORDER_TYPE Type;
  double Price;
  double PriceCurrent;
  long PositionID;
  double Lots;

  long StringToTime2( const string Str ) const
  {
    return(((datetime)Str * 1000 + (int)::StringSubstr(Str, ::StringLen(Str) - 6, 3)));
  }

  bool Set( string Str )
  {
    ::StringReplace(Str, ",", ".");

    string StrArray[10];

    const bool Res = (::StringSplit(Str, '\t', StrArray) == ::ArraySize(StrArray));

    if (Res)
    {
      this.Ticket = (long)StrArray[0];
      this.Symb = StrArray[1];
      this.State = (StrArray[2] == "FILL") ? ORDER_STATE_FILLED : ORDER_STATE_REJECTED;

      this.TimeSetup = this.StringToTime2(StrArray[3]);
      this.TimeDone = this.StringToTime2(StrArray[4]);

      this.Type = (StrArray[5] == "BUY") ? ORDER_TYPE_BUY : ORDER_TYPE_SELL;

      this.Price = (double)StrArray[6];
      this.PriceCurrent = (double)StrArray[7];

      this.PositionID = (long)StrArray[8];
      this.Lots = (double)StrArray[9];
    }

    return(Res);
  }

  string ToString( void ) const
  {
    const int digits = (int)SymbolInfoInteger(this.Symb[], SYMBOL_DIGITS);

    return((string)this.Ticket + " " + this.Symb[] + " " + ::EnumToString(this.State) + " " +
           ::TimeToString(this.TimeSetup) + " " + ::TimeToString(this.TimeDone) +  " " +
           ::EnumToString(this.Type) + " " + ::DoubleToString(this.Price, digits) + " " +
           ::DoubleToString(this.PriceCurrent, digits) + " " + (string)PositionID + " " + DoubleToString(this.Lots, 2));
  }

  bool operator ==( const ORDER &Order ) const
  {
    return((this.Type == Order.Type) && (this.Price == Order.Price) && (this.Symb[] == Order.Symb[]));
  }

  long GetLength( void ) const
  {
    return(this.TimeDone - this.TimeSetup);
  }
};

int GetSymbols( const ORDER &Orders[], string &Symbols[] )
{
  ArrayFree(Symbols);

  for (int i = ArraySize(Orders) - 1; i >= 0; i--)
  {
    const string Symb = Orders[i].Symb[];
    const int Size = ArraySize(Symbols);

    int k = Size;

    while ((--k >= 0) && (Symbols[k] != Symb))
      ;

    if (k < 0)
      Symbols[ArrayResize(Symbols, Size + 1) - 1] = Symb;
  }

  return(ArraySize(Symbols));
}

int Filter( const ORDER &Orders[], ORDER &FilterOrders[], const string Symb = NULL, const long From = 0 )
{
  const int Size = ArrayResize(FilterOrders, ArraySize(Orders));
  int Amount = 0;

  for (int i = 0; i < Size; i++)
    if ((Orders[i].TimeDone >= From) && ((Symb == NULL) || (Orders[i].Symb[] == Symb)))
      FilterOrders[Amount++] = Orders[i];

  return(ArrayResize(FilterOrders, Amount));
}

int GetAmountState( const ORDER &Orders[], const ENUM_ORDER_STATE State )
{
  int Res = 0;

  for (int i = ArraySize(Orders) - 1; i >= 0; i--)
    if (Orders[i].State == State)
      Res++;

  return(Res);
}

int GetLengthState( const ORDER &Orders[], const ENUM_ORDER_STATE State )
{
  int Res = 0;

  for (int i = ArraySize(Orders) - 1; i >= 0; i--)
    if (Orders[i].State == State)
      Res += (int)Orders[i].GetLength();

  return(Res);
}

int Filter( const ORDER &Orders[], ORDER &FilterOrders[], const ENUM_ORDER_STATE State )
{
  const int Size = ArrayResize(FilterOrders, ArraySize(Orders));
  int Amount = 0;

  long TimeSetup = 0;

  for (int i = 0; i < Size; i++)
    if (((State == ORDER_STATE_PLACED) || (Orders[i].State == State)) && (Orders[i].TimeSetup != TimeSetup))
    {
      FilterOrders[Amount++] = Orders[i];

      TimeSetup = Orders[i].TimeSetup;
    }

  return(ArrayResize(FilterOrders, Amount));
}

int CreateFillOrders( const ORDER &Orders[], ORDER &FillOrders[], const int MinCount = 2 )
{
  ORDER FilterOrders[];
//  const int Size1 = Filter(Orders, FilterOrders, ORDER_STATE_FILLED);
  const int Size1 = Filter(Orders, FilterOrders, ORDER_STATE_PLACED);

  const int Size2 = ArrayResize(FillOrders, ArraySize(Orders));

  int Amount = 0;
  int j = 0;

  for (int i = 0; i < Size1; i++)
  {
    const long TimeSetup = FilterOrders[i].TimeSetup;

    while (((j < Size2) && (Orders[j].TimeSetup != TimeSetup)))
      j++;

    if ((j < Size2) && (Orders[j].TimeSetup == TimeSetup))
    {
      int Count = 0;

      while ((j < Size2) && (Orders[j].TimeSetup == TimeSetup))
      {
        FillOrders[Amount++] = Orders[j++];

        Count++;
      }

      if (Count < MinCount)
        Amount -= Count;
    }
  }

  return(ArrayResize(FillOrders, Amount));
}

struct STAT
{
  int Amount;
  int Filled;

  long FillTime;
  long RejectTime;

  STAT( void ) : Amount(0), Filled(0), FillTime(0), RejectTime(0)
  {
  }

  void Set( const ORDER &Order )
  {
    this.Amount++;

    if (Order.State == ORDER_STATE_FILLED)
    {
      FillTime += Order.GetLength();

      this.Filled++;
    }
    else
      RejectTime += Order.GetLength();

    return;
  }

  double GetFillRate( void ) const
  {
    return(this.Amount ? (double)this.Filled / this.Amount : 0);
  }

  string ToString( void ) const
  {
    return("FillRate = " + ::DoubleToString(100.0 * this.GetFillRate(), 2) +
           "% (" + (string)this.Filled + "/" + (string)this.Amount + ")" +
           " FillTime = " + (string)(this.Filled ? this.FillTime / this.Filled : 0) +
           " ms., RejectTime = " + (string)((this.Amount != this.Filled) ? this.RejectTime / (this.Amount - this.Filled) : 0) + " ms.");
  }

};

int StatFillOrders( const ORDER &Orders[], STAT &Stat[], const int MinCount = 2 )
{
  ArrayFree(Stat);

  ORDER FillOrders[];
  const int Size = CreateFillOrders(Orders, FillOrders, MinCount);

  int i = 0;

  while (i < Size)
  {
    const long TimeSetup = FillOrders[i].TimeSetup;

    for (int Count = 0; (i < Size) && (FillOrders[i].TimeSetup == TimeSetup); i++, Count++)
    {
      if (Count >= ArraySize(Stat))
        ArrayResize(Stat, Count + 1);

      Stat[Count].Set(FillOrders[i]);
    }
  }

  return(ArraySize(Stat));
}

void CalcStat( const ORDER &Orders[], const int MinCount = 2 )
{
  const int Size = ArraySize(Orders);

  const int AmountFill = GetAmountState(Orders, ORDER_STATE_FILLED);
  const int AmountReject = Size - AmountFill;

  const int LengthFill = AmountFill ? GetLengthState(Orders, ORDER_STATE_FILLED) / AmountFill : 0;
  const int LengthReject = AmountReject ? GetLengthState(Orders, ORDER_STATE_REJECTED) / AmountReject : 0;

  Print(Orders[0].Symb[] + " FillRate = " + DoubleToString(100.0 * AmountFill / Size, 2) + "% (" + (string)AmountFill + "/" + (string)Size + ")\n" +
        Orders[0].Symb[] + " Average FillTime = " + (string)LengthFill + " ms., Average RejectTime = " + (string)LengthReject + " ms.");

  for (int Count = 1; Count <= MinCount; Count++)
  {
    STAT Stat[];

    const int Size2 = MathMin(StatFillOrders(Orders, Stat, Count), 5);

    Print("\n" + Orders[0].Symb[] + " .  ,   = " + (string)Count + ":");

    for (int i = 0; i < Size2; i++)
      if (/*(Stat[i].Amount > 20) &&*/ (!i || Stat[i].GetFillRate() < Stat[i - 1].GetFillRate()))
        Print((string)(i + 1) + ": " + Stat[i].ToString());
  }
}

void CalcStat2( const ORDER &Orders[] )
{
  STAT Stat[];

  const int Size = MathMin(GetStat2(Orders, Stat), 5);

  Print("\n" + Orders[0].Symb[] + "     ,     :");

  for (int i = 0; i < Size; i++)
    Print((string)(i + 1) + ": " + Stat[i].ToString());
}


//  ,        ,     .
int GetStat2( const ORDER &Orders[], STAT &Stat[] )
{
  ArrayFree(Stat);

  const int Size = ArraySize(Orders);

  for (int i = 0; i < Size;)
  {
    const long TimeSetup = Orders[i++].TimeSetup;

    int Count = 1;

    for (; (i < Size) && (Orders[i].TimeSetup == TimeSetup); i++, Count++)
      ;

    if (Count >= ArraySize(Stat))
      ArrayResize(Stat, Count);

    Stat[Count - 1].Set(Orders[i - Count]);
  }

  return(ArraySize(Stat));
}

void OnStart()
{
  uchar Data[];

  FileLoad(inFileName, Data);

  string StrArray[];
  ORDER Orders[];

  const int Size = ArrayResize(Orders, StringSplit(CharArrayToString(Data), '\n', StrArray) - 1);

  for (int i = 0; i < Size; i++)
    Orders[i].Set(StrArray[i + 1]);

  string Symbols[];

  const int Amount = GetSymbols(Orders, Symbols);

  for (int i = 0; i < Amount; i++)
  {
    ORDER FilterOrders[];

    if (Filter(Orders, FilterOrders, Symbols[i]) > 300)
    {
      CalcStat(FilterOrders, inMinCount);
      CalcStat2(FilterOrders);

      Print("");
    }
  }
}
